"""
Custom defined Click commands/options/etc.
"""

import click

from griffon.exceptions import GriffonUsageError
from griffon.helpers import Style


class ListParamType(click.ParamType):
    """Custom comma-separated list type"""

    name = "list"

    def convert(self, value, param, ctx):
        if value is None:
            return []
        return value.split(",")


class BaseGroupParameter:
    """
    Custom base parameter which handles:
        * mutually exclusive options
        * required group options (one of the group is required)
    """

    def handle_parse_result(self, ctx, opts, args):
        if not ctx.resilient_parsing:
            if self.mutually_exclusive_group:
                if opts.get(self.name) is None:
                    pass  # skip check for not supplied click.Arguments
                elif self.name in opts and any(
                    opt in opts
                    for opt in self.mutually_exclusive_group
                    if opts.get(opt) is not None
                ):
                    raise GriffonUsageError(
                        (
                            f"{Style.BOLD}{self.name} cannot be used with "
                            f"{', '.join(self.mutually_exclusive_group)}.{Style.RESET}"
                        ),
                        ctx=ctx,
                    )

            if self.required_group:
                group_set = set(
                    opt for opt in opts if opt in self.required_group and opts.get(opt) is not None
                )
                if not any(group_set):
                    raise GriffonUsageError(
                        f"{Style.BOLD}At least one of {', '.join(self.required_group)} "
                        f"is required.{Style.RESET}",
                        ctx=ctx,
                    )

        return super().handle_parse_result(ctx, opts, args)


class GroupOption(BaseGroupParameter, click.Option):
    """Custom Option with BaseGroupParameter functionality"""

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive_group = set(kwargs.pop("mutually_exclusive_group", []))
        self.required_group = set(kwargs.pop("required_group", []))

        if self.mutually_exclusive_group:
            mutually_exclusive_str = ", ".join(self.mutually_exclusive_group)
            kwargs["help"] = kwargs.get("help", "") + (
                f", this argument is mutually exclusive "
                f"with arguments: {Style.BOLD}[{mutually_exclusive_str}]{Style.RESET}"
            )

        if self.required_group:
            required_str = ", ".join(self.required_group)
            kwargs["help"] = kwargs.get("help", "") + (
                f", at least one of these arguments: "
                f"{Style.BOLD}[{required_str}]{Style.RESET} is required."
            )

        super().__init__(*args, **kwargs)


class GroupArgument(BaseGroupParameter, click.Argument):
    """Custom Argument with BaseGroupParameter functionality"""

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive_group = set(kwargs.pop("mutually_exclusive_group", []))
        self.required_group = set(kwargs.pop("required_group", []))

        super().__init__(*args, **kwargs)



"""
Helpers for direct usage or debbuging
"""
import json
from enum import Enum
from typing import Callable, Optional, Type, Union

from component_registry_bindings.bindings.python_client.types import (
    ComponentRegistryModel,
)
from osidb_bindings.bindings.python_client.types import OSIDBModel


def debug_data_dump(filename: str, data, transform_fn: Optional[Callable] = None):
    """
    Debugging utility to avoid heavy HTTP data transfers.
    Serializes the data into JSON and dumps that into a
    specified file.

    transform function can be used for postprocessing the data after
    the basic dict serialization

    """
    if isinstance(data, list):
        json_data = [
            item.to_dict() if isinstance(item, (ComponentRegistryModel, OSIDBModel)) else item
            for item in data
        ]

        if transform_fn is not None:
            json_data = list(map(transform_fn, json_data))

    else:
        json_data = data

    with open(filename, "w") as fp:
        json.dump(json_data, fp)


def debug_data_load(
    filename: str,
    model: Optional[Union[Type[ComponentRegistryModel], Type[OSIDBModel]]] = None,
    transform_fn: Optional[Callable] = None,
):
    """
    Debugging utility to avoid heavy HTTP data transfers.
    Loads the data from specified JSON file and transforms them
    back into internal models.

    transform function can be used for preprocessing the data before
    the internal model transformation
    """
    with open(filename) as fp:
        json_data = json.load(fp)

    if isinstance(json_data, list):
        if transform_fn is not None:
            json_data = list(map(transform_fn, json_data))

        data = [model.from_dict(item) for item in json_data]  # type: ignore

    else:
        data = json_data
    return data


class Color(Enum):
    """
    Helper enum for text color formatting for anything which
    cannot be rendered using rich
    """

    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    GREY = "\033[90m"
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"
    RESET = "\033[0m"  # Reset to default color and style

    def __str__(self):
        return str(self.value)


class Style(Enum):
    """
    Helper enum for text style formatting for anything which
    cannot be rendered using rich
    """

    BOLD = "\033[1m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"
    STRIKE = "\033[9m"
    RESET = "\033[0m"  # Reset to default style

    def __str__(self):
        return str(self.value)



"""
minimal example of the Griffon plugin
"""

import click


@click.group(help="Example plugin")
@click.pass_context
def plugins(ctx):
    """Example plugin"""
    pass


@plugins.command()
def test():
    """Test command"""
    click.echo("Command 'test' of the 'example' plugin executed")


@plugins.command()
@click.option("-m", "--message", "message", help="Message to print out.")
def test_with_option(message):
    """Test with option command"""
    click.echo("Command 'test_with_option' of the 'example' plugin executed")
    click.echo(message)



"""
    read only queries

"""
import copy
import logging
import multiprocessing
import re
from functools import partial
from typing import Any, Dict, List

from component_registry_bindings.bindings.python_client.models import Component

from griffon import (
    COMMUNITY_COMPONENTS_SERVER_URL,
    CORGI_SERVER_URL,
    OSIDB_SERVER_URL,
    CommunityComponentService,
    CorgiService,
    OSIDBService,
)

logger = logging.getLogger("griffon")


class product_stream_summary:
    """retrieve product_stream summary"""

    name = "product_stream_summary"
    description = "retrieve product_stream summary"
    allowed_params = [
        "strict_name_search",
        "all",
        "product_stream_name",
        "ofuri",
        "verbose",
        "regex_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.product_stream_name = self.params.get("product_stream_name")
        self.ofuri = self.params.get("ofuri")
        self.strict_name_search = self.params.get("strict_name_search", None)
        self.all = self.params.get("all", None)
        self.regex_name_search = self.params.get("regex_name_search")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        cond = {}

        product_stream_name = (
            re.escape(self.product_stream_name)
            if not self.regex_name_search
            else self.product_stream_name
        )

        if self.ofuri:
            cond["ofuri"] = self.ofuri
        elif not self.strict_name_search:
            if not self.all:
                cond["re_name"] = product_stream_name
        else:
            if not self.all:
                cond["name"] = product_stream_name

        # TODO - corgi bindings need to support ofuri in product_streams
        product_streams = self.corgi_session.product_streams.retrieve_list(
            **cond,
            limit=400,
            include_fields="link,ofuri,name,products,product_versions,brew_tags,manifest",
        )

        results = []
        if product_streams.results:
            for ps in product_streams.results:
                result = {
                    "link": ps.link,
                    "ofuri": ps.ofuri,
                    "name": ps.name,
                    "product": ps.products[0]["name"],
                    "product_version": str([pv["name"] for pv in ps.product_versions]),
                    "brew_tags": [brew_tag for brew_tag in ps.brew_tags.to_dict().keys()],
                    # "build_count": ps.build_count,
                    "manifest_link": ps.manifest,
                    "latest_components_link": f"{CORGI_SERVER_URL}/api/v1/components?ofuri={ps.ofuri}&view=summary",  # noqa
                    "all_components_link": f"{CORGI_SERVER_URL}/api/v1/components?product_streams={ps.ofuri}&include_fields=link,name,purl",  # noqa
                }
                results.append(result)
        else:
            if "ofuri" in product_streams:
                result = {
                    "link": product_streams["link"],
                    "ofuri": product_streams["ofuri"],
                    "name": product_streams["name"],
                    "product": product_streams["products"][0]["name"],
                    "product_version": product_streams["product_versions"][0]["name"],
                    "brew_tags": "",
                    "manifest_link": product_streams["manifest"],
                    "latest_components_link": f"{CORGI_SERVER_URL}/api/v1/components?ofuri={product_streams['ofuri']}&view=summary",  # noqa
                    "all_components_link": f"{CORGI_SERVER_URL}/api/v1/components?product_streams={product_streams['ofuri']}&include_fields=link,name,purl",  # noqa
                }
                results.append(result)
            else:
                logger.warning("No such product stream")
        return results


class products_versions_affected_by_specific_cve_query:
    """Given a specific CVE ID, what products are affected?"""

    name = "product-versions-affected-by-specific-cve"
    description = "Given a specific CVE ID, what product versions are affected?"
    allowed_params = ["cve_id"]

    def __init__(self, params: dict) -> None:
        self.osidb_session = OSIDBService.create_session()
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.cve_id = self.params.get("cve_id")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")
        self.component_type = self.params.get("component_type")
        self.namespace = self.params.get("namespace")

    def execute(self, status=None) -> dict:
        cond = {}
        if self.affectedness:
            cond["affectedness"] = self.affectedness
        if self.affect_resolution:
            cond["resolution"] = self.affect_resolution
        if self.affect_impact:
            cond["impact"] = self.affect_impact

        component_cond = {}
        if self.namespace:
            component_cond["namespace"] = self.namespace
        if self.component_type:
            component_cond["type"] = self.component_type

        flaw = self.osidb_session.flaws.retrieve(self.cve_id)
        affects = flaw.affects
        results = list()
        product_versions = set()
        product_streams = set()
        for affect in affects:
            components = self.corgi_session.components.retrieve_list_iterator_async(
                name=affect.ps_component,
                latest_components_by_streams="True",
                include_fields="product_streams.name,product_versions.name",
            )
            for c in components:
                results.append(c.to_dict())
        for c in results:
            for ps in c["product_streams"]:
                product_streams.add(ps["name"])
            for pv in c["product_versions"]:
                product_versions.add(ps["name"])
        return {
            "link": f"{OSIDB_SERVER_URL}/osidb/api/v1/flaws/{flaw.cve_id}",
            "cve_id": flaw.cve_id,
            "title": flaw.title,
            "description": flaw.description,
            "product_versions": sorted(list(product_versions)),
            "product_streams": sorted(list(product_streams)),
            # "components": results,
        }


class products_containing_specific_component_query:
    """What products contain a specific component?"""

    name = "products_containing_specific_component_query"
    description = "What products contain a specific component?"
    allowed_params = [
        "component_name",
        "purl",
        "arch",
        "namespace",
        "component_type",
        "strict_name_search",
        "affect_mode",
        "search_latest",
        "search_provides",
        "search_upstreams",
        "search_all",
        "search_all_roots",
        "search_all_upstreams",
        "search_related_url",
        "search_community",
        "search_upstreams",
        "filter_rh_naming",
        "search_redhat",
        "no_community",
        "no_middleware",
        "no_upstream_affects",
        "include_inactive_product_streams",
        "include_product_stream_excluded_components",
        "output_type_filter",
        "regex_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.purl = self.params.get("purl")

    def execute(self, status=None) -> dict:
        c = self.corgi_session.components.retrieve_list(
            purl=self.purl,
        )
        return c["product_streams"]


def async_retrieve_sources(self, purl):
    params = {
        "limit": 200,
        "root_components": "True",
        "released_components": "True",
        "provides": purl,
        "include_fields": "type,nvr,purl,name,namespace,download_url,related_url",
    }
    # TODO: remove cnt and max_result and cnt > 10000 on next stage-> prod deployment
    cnt = self.components.count(**params)
    if cnt == 0:
        return []
    try:
        if cnt > 10000:
            return list(self.components.retrieve_list_iterator_async(max_results=5000, **params))
        return list(self.components.retrieve_list_iterator_async(max_results=10000, **params))
    except Exception as e:
        logger.warning(f"{type(e).__name__} - problem retrieving all of {purl} {cnt} sources.")
        return []


def async_retrieve_upstreams(self, purl):
    params = {
        "limit": 200,
        "root_components": "True",
        "upstreams": purl,
        "include_fields": "type,nvr,purl,name,namespace,download_url,related_url",
    }
    # TODO: remove cnt and max_result and cnt > 10000 on next stage-> prod deployment
    cnt = self.components.count(**params)
    if cnt == 0:
        return []
    try:
        if cnt > 10000:
            return list(self.components.retrieve_list_iterator_async(max_results=5000, **params))
        return list(self.components.retrieve_list_iterator_async(max_results=10000, **params))
    except Exception as e:
        logger.warning(f"{type(e).__name__} - problem retrieving all of {purl} {cnt} upstreams.")
        return []


def process_component(session, c):
    """perform any neccessary sub retrievals."""
    # only process provided components
    if c.software_build is None:
        c.sources = async_retrieve_sources(session, c.purl)
        c.upstreams = async_retrieve_upstreams(session, c.purl)
    return c


class products_containing_component_query:
    """What products contain a component?"""

    name = "products_containing_component_query"
    description = "What products contain a component?"
    allowed_params = [
        "component_name",
        "purl",
        "arch",
        "namespace",
        "component_type",
        "strict_name_search",
        "search_latest",
        "search_provides",
        "search_upstreams",
        "search_all",
        "search_all_roots",
        "search_all_upstreams",
        "search_related_url",
        "search_redhat",
        "search_community",
        "search_upstreams",
        "filter_rh_naming",
        "no_community",
        "no_middleware",
        "no_upstream_affects",
        "include_inactive_product_streams",
        "include_product_stream_excluded_components",
        "output_type_filter",
        "regex_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.component_name = self.params.get("component_name", "")
        self.component_type = self.params.get("component_type", "")
        self.strict_name_search = self.params.get("strict_name_search")
        self.search_deps = self.params.get("search_deps")
        self.ns = self.params.get("namespace")
        self.search_latest = self.params.get("search_latest")
        self.search_all = self.params.get("search_all")
        self.search_all_roots = self.params.get("search_all_roots")
        self.search_related_url = self.params.get("search_related_url")
        self.search_redhat = self.params.get("search_redhat")
        self.search_community = self.params.get("search_community")
        self.search_all_upstreams = self.params.get("search_all_upstreams")
        self.filter_rh_naming = self.params.get("filter_rh_naming")
        self.no_community = self.params.get("no_community")
        self.search_provides = self.params.get("search_provides")
        self.search_upstreams = self.params.get("search_upstreams")
        self.regex_name_search = self.params.get("regex_name_search")
        if not self.no_community:
            self.community_session = CommunityComponentService.create_session()
        self.include_inactive_product_streams = self.params.get("include_inactive_product_streams")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        status.update("searching component-registry.")
        results = []
        params = {
            "limit": 50,
            "include_fields": "purl,type,name,related_url,namespace,software_build,nvr,release,version,arch,product_streams.product_versions,product_streams.name,product_streams.ofuri,product_streams.active,product_streams.exclude_components",  # noqa
        }

        component_name = (
            re.escape(self.component_name) if not self.regex_name_search else self.component_name
        )

        if self.search_latest:
            search_latest_params = copy.deepcopy(params)
            if not (self.strict_name_search):
                search_latest_params["re_name"] = component_name
            else:
                search_latest_params["name"] = component_name
            if self.ns:
                search_latest_params["namespace"] = self.ns
            if not (self.include_inactive_product_streams):
                search_latest_params["active_streams"] = "True"
            search_latest_params["root_components"] = "True"
            search_latest_params["latest_components_by_streams"] = "True"
            status.update("searching latest root component(s).")
            latest_components_cnt = self.corgi_session.components.count(**search_latest_params)
            status.update(f"found {latest_components_cnt} latest component(s).")
            latest_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_latest_params
            )
            status.update(f"found {latest_components_cnt} latest root component(s).")  # noqa
            for c in latest_components:
                results.append(c)

            if not self.no_community:
                status.update("searching latest community root component(s).")
                community_component_cnt = self.community_session.components.count(
                    **search_latest_params
                )
                status.update(
                    f"found {community_component_cnt} latest community root component(s)."  # noqa
                )
                latest_community_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_latest_params
                    )
                )
                status.update(
                    f"found {community_component_cnt} latest community root component(s), retrieving sources & upstreams."  # noqa
                )
                for c in latest_community_components:
                    results.append(c)

        if self.search_provides:
            search_provides_params = copy.deepcopy(params)
            if not (self.strict_name_search):
                search_provides_params["re_provides_name"] = component_name
            else:
                search_provides_params["provides_name"] = component_name
            if self.ns:
                search_provides_params["namespace"] = self.ns
            if not (self.include_inactive_product_streams):
                search_provides_params["active_streams"] = "True"
            search_provides_params["latest_components_by_streams"] = "True"
            status.update("searching latest provided child component(s).")
            latest_components_cnt = self.corgi_session.components.count(**search_provides_params)
            status.update(f"found {latest_components_cnt} latest provides component(s).")
            latest_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_provides_params
            )
            status.update(
                f"found {latest_components_cnt} latest provides child component(s), retrieving sources & upstreams."  # noqa
            )
            for c in latest_components:
                results.append(c)

            if not self.no_community:
                status.update("searching latest community provided child component(s).")
                community_component_cnt = self.community_session.components.count(
                    **search_provides_params
                )
                status.update(
                    f"found {community_component_cnt} latest community provided child component(s)."  # noqa
                )
                latest_community_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_provides_params
                    )
                )
                status.update(
                    f"found {community_component_cnt} latest community provided child component(s), retrieving sources & upstreams."  # noqa
                )
                for c in latest_community_components:
                    results.append(c)

        if self.search_upstreams:
            search_upstreams_params = copy.deepcopy(params)
            search_upstreams_params["latest_components_by_streams"] = "True"
            if not (self.strict_name_search):
                search_upstreams_params["re_upstreams_name"] = component_name
            else:
                search_upstreams_params["upstreams_name"] = component_name
            if self.ns:
                search_upstreams_params["namespace"] = self.ns
            if not (self.include_inactive_product_streams):
                search_upstreams_params["active_streams"] = "True"
            search_upstreams_params["released_components"] = "True"
            search_upstreams_params["latest_components_by_streams"] = "True"
            status.update("searching latest upstreams child component(s).")
            latest_components_cnt = self.corgi_session.components.count(**search_upstreams_params)
            status.update(f"found {latest_components_cnt} latest component(s).")
            latest_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_upstreams_params
            )
            with multiprocessing.Pool() as pool:
                status.update(
                    f"found {latest_components_cnt} latest upstreams child component(s), retrieving sources & upstreams."  # noqa
                )
                for processed_component in pool.map(
                    partial(process_component, self.corgi_session), latest_components
                ):
                    results.append(processed_component)
            if not self.no_community:
                status.update("searching latest community upstreams child component(s).")
                community_component_cnt = self.community_session.components.count(
                    **search_upstreams_params
                )
                status.update(
                    f"found {community_component_cnt} latest community upstreams child component(s)."  # noqa
                )
                latest_community_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_upstreams_params
                    )
                )
                with multiprocessing.Pool() as pool:
                    status.update(
                        f"found {community_component_cnt} latest community provided child component(s), retrieving sources & upstreams."  # noqa
                    )
                    for processed_component in pool.map(
                        partial(process_component, self.community_session),
                        latest_community_components,
                    ):
                        results.append(processed_component)

        if self.search_related_url:
            search_related_url_params = copy.deepcopy(params)
            # Note: related_url filter has no concept of strict
            search_related_url_params["related_url"] = component_name
            if self.ns:
                search_related_url_params["namespace"] = self.ns
            if self.component_type:
                search_related_url_params["type"] = self.component_type
            if not (self.include_inactive_product_streams):
                search_related_url_params["active_streams"] = "True"
            search_related_url_params["released_components"] = "True"
            related_url_components_cnt = self.corgi_session.components.count(
                **search_related_url_params
            )
            status.update(f"found {related_url_components_cnt} related url component(s).")
            related_url_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_related_url_params
            )
            with multiprocessing.Pool() as pool:
                status.update(
                    f"found {related_url_components_cnt} related url component(s), retrieving sources & upstreams."  # noqa
                )
                for processed_component in pool.map(
                    partial(process_component, self.corgi_session), related_url_components
                ):
                    results.append(processed_component)
            if not self.no_community:
                latest_community_url_components_cnt = self.community_session.components.count(
                    **search_related_url_params
                )
                status.update(
                    f"found {latest_community_url_components_cnt} related url community component(s)."  # noqa
                )
                latest_community_url_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_related_url_params
                    )
                )
                with multiprocessing.Pool() as pool:
                    status.update(
                        f"found {latest_community_url_components_cnt} related url community component(s), retrieving sources & upstreams."  # noqa
                    )
                    for processed_component in pool.map(
                        partial(process_component, self.community_session),
                        latest_community_url_components,
                    ):
                        results.append(processed_component)

        if self.search_all:
            search_all_params = copy.deepcopy(params)
            if not (self.strict_name_search):
                search_all_params["re_name"] = component_name
            else:
                search_all_params["name"] = component_name
            if self.component_type:
                search_all_params["type"] = self.component_type
            if self.ns:
                search_all_params["namespace"] = self.ns
            if not (self.include_inactive_product_streams):
                search_all_params["active_streams"] = "True"
            search_all_params["released_components"] = "True"
            all_components_cnt = self.corgi_session.components.count(**search_all_params)
            status.update(f"found {all_components_cnt} all component(s).")
            # TODO: remove max_results
            all_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_all_params, max_results=10000
            )
            status.update(f"found {all_components_cnt} all component(s).")
            with multiprocessing.Pool() as pool:
                status.update(
                    f"found {all_components_cnt} all component(s), retrieving sources & upstreams."  # noqa
                )
                for processed_component in pool.map(
                    partial(process_component, self.corgi_session), all_components
                ):
                    results.append(processed_component)

            if not self.no_community:
                all_community_components_cnt = self.community_session.components.count(
                    **search_all_params
                )
                status.update(
                    f"found {all_community_components_cnt} community all component(s)."  # noqa
                )
                # TODO: remove max_results
                all_community_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_all_params, max_results=10000
                    )
                )
                with multiprocessing.Pool() as pool:
                    status.update(
                        f"found {all_community_components_cnt} community all component(s), retrieving sources & upstreams."  # noqa
                    )
                    for processed_component in pool.map(
                        partial(process_component, self.community_session),
                        all_community_components,
                    ):
                        results.append(processed_component)

        if self.search_all_roots:
            search_all_roots_params = copy.deepcopy(params)
            search_all_roots_params["root_components"] = "True"
            if not (self.strict_name_search):
                search_all_roots_params["re_name"] = component_name
            else:
                search_all_roots_params["name"] = component_name
            if self.ns:
                search_all_roots_params["namespace"] = self.ns
            if not (self.include_inactive_product_streams):
                search_all_roots_params["active_streams"] = "True"
            search_all_roots_params["released_components"] = "True"
            all_src_components_cnt = self.corgi_session.components.count(**search_all_roots_params)
            status.update(f"found {all_src_components_cnt} all root component(s).")
            all_src_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_all_roots_params
            )
            for c in all_src_components:
                results.append(c)
            if not self.no_community:
                all_src_community_components_cnt = self.community_session.components.count(
                    **search_all_roots_params
                )
                all_src_community_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_all_roots_params
                    )
                )
                status.update(
                    f"found {all_src_community_components_cnt} community all root component(s)."  # noqa
                )
                for c in all_src_community_components:
                    results.append(c)

        if self.search_all_upstreams:
            search_all_upstreams_params = copy.deepcopy(params)
            search_all_upstreams_params["namespace"] = "UPSTREAM"
            if not (self.strict_name_search):
                search_all_upstreams_params["re_name"] = component_name
            else:
                search_all_upstreams_params["name"] = component_name
            if self.component_type:
                search_all_upstreams_params["type"] = self.component_type
            if not (self.include_inactive_product_streams):
                search_all_upstreams_params["active_streams"] = "True"
            upstream_components_cnt = self.corgi_session.components.count(
                **search_all_upstreams_params
            )
            status.update(f"found {upstream_components_cnt} upstream component(s).")
            upstream_components = self.corgi_session.components.retrieve_list_iterator_async(
                **search_all_upstreams_params
            )
            with multiprocessing.Pool() as pool:
                status.update(f"found {upstream_components_cnt} upstream component(s).")
                for processed_component in pool.map(
                    partial(process_component, self.corgi_session), upstream_components
                ):
                    results.append(processed_component)
            if not self.no_community:
                commmunity_upstream_components_cnt = self.community_session.components.count(
                    **search_all_upstreams_params
                )
                status.update(
                    f"found {commmunity_upstream_components_cnt} community upstream component(s)."  # noqa
                )
                commmunity_upstream_components = (
                    self.community_session.components.retrieve_list_iterator_async(
                        **search_all_upstreams_params
                    )
                )
                with multiprocessing.Pool() as pool:
                    status.update(
                        f"found {commmunity_upstream_components_cnt} community upstream component(s), retrieving sources & upstreams."  # noqa
                    )
                    for processed_component in pool.map(
                        partial(process_component, self.community_session),
                        commmunity_upstream_components,
                    ):
                        results.append(processed_component)

        if self.filter_rh_naming:
            flags = re.IGNORECASE
            patterns = [
                # binutils
                re.compile(
                    f"(devtoolset\\-[0-9]+\\-|mingw\\-|gcc\\-toolset\\-[0-9]+\\-)?{component_name}[0-9\\.]*$",  # noqa
                    flags=flags,
                ),
                # compat-* style
                re.compile(f"(compat\\-)?{component_name}[0-9\\.]*(\\-[0-9]+)?$", flags=flags),
                # kernel
                re.compile(f"^{component_name}(\\-rt)?$", flags=flags),
                # qemu
                re.compile(f"^{component_name}(\\-kvm(\\-rhev|\\-ma)?)?$", flags=flags),
                # webkit
                re.compile(f"^{component_name}([0-9])?(gtk)?([0-9])?$", flags=flags),
            ]

            filtered_results = []
            for result in results:
                is_matched = False
                for p in patterns:
                    if is_matched:
                        break
                    if type(result) == Component:
                        m = p.match(result.name)
                        if m:
                            filtered_results.append(result)
                            is_matched = True
                            break
                    else:
                        m = p.match(result["name"])
                        if m:
                            filtered_results.append(result)
                            is_matched = True
                            break

            results = filtered_results

        if self.search_community:
            search_community_params = copy.deepcopy(params)
            if not (self.strict_name_search):
                search_community_params["re_name"] = component_name
            else:
                search_community_params["name"] = component_name
            if self.ns:
                search_community_params["namespace"] = self.ns
            if not (self.include_inactive_product_streams):
                search_community_params["active_streams"] = "True"
            all_community_components_cnt = self.community_session.components.count(
                **search_community_params
            )
            status.update(
                f"found {all_community_components_cnt} community all component(s)."  # noqa
            )
            all_community_components = (
                self.community_session.components.retrieve_list_iterator_async(
                    **search_community_params
                )
            )
            with multiprocessing.Pool() as pool:
                status.update(
                    f"found {all_community_components_cnt} community all component(s), retrieving sources & upstreams."  # noqa
                )
                for processed_component in pool.map(
                    partial(process_component, self.community_session), all_community_components
                ):
                    results.append(processed_component)

        return results


class components_containing_specific_component_query:
    """What components contain a specific component?"""

    name = "components_containing_specific_component_query"
    description = "What components contain a specific component?"
    allowed_params = [
        "component_re_name",
        "component_name",
        "purl",
        "component_type",
        "component_version",
        "component_arch",
        "namespace",
        "strict_name_search",
        "verbose",
    ]

    def __init__(self, params: dict):
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.purl = self.params.get("purl")

    def execute(self, status=None) -> dict:
        if self.purl:
            c = self.corgi_session.components.retrieve_list(
                purl=self.purl,
            )
            component_type = self.params["component_type"]
            sources = c["sources"]
            if component_type:
                sources = [source for source in sources if component_type.lower() in source["purl"]]
        return {
            "link": c["link"],
            "type": c["type"],
            "name": c["name"],
            "version": c["version"],
            "purl": c["purl"],
            "sources": sources,
        }


class components_containing_component_query:
    """What components contain a component?"""

    name = "components_containing_component_query"
    description = "What components contain a component?"
    allowed_params = [
        "component_name",
        "purl",
        "component_type",
        "component_version",
        "component_arch",
        "namespace",
        "strict_name_search",
        "verbose",
        "regex_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.component_name = self.params.get("component_name")
        self.component_type = self.params.get("component_type")
        self.component_version = self.params.get("component_version")
        self.component_arch = self.params.get("component_arch")
        self.namespace = self.params.get("namespace")
        self.strict_name_search = self.params.get("strict_name_search")
        self.regex_name_search = self.params.get("regex_name_search")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        component_name = (
            re.escape(self.component_name) if not self.regex_name_search else self.component_name
        )

        cond = {}
        if not self.strict_name_search:
            cond["re_name"] = component_name
        else:
            cond["name"] = component_name
        if self.component_version:
            cond["version"] = self.component_version
        if self.component_arch:
            cond["arch"] = self.component_arch
        if self.namespace:
            cond["namespace"] = self.namespace

        components = self.corgi_session.components.retrieve_list_iterator_async(**cond)
        results = []
        for c in components:
            sources = [{"link": source["link"], "purl": source["purl"]} for source in c.sources]
            if self.component_type:
                sources = [
                    source for source in sources if self.component_type.lower() in source["purl"]
                ]
            results.append(
                {
                    "link": c.link,
                    "type": c.type,
                    "name": c.name,
                    "version": c.version,
                    "nvr": c.nvr,
                    "arch": c.arch,
                    "download_url": c.download_url,
                    "related_url": c.related_url,
                    "purl": c.purl,
                    "sources": sources,
                }
            )
        return results


class components_affected_by_specific_cve_query:
    """Given a specific CVE ID, what components are affected?"""

    name = "components_affected_by_specific_cve_query"
    description = "Given a CVE ID, what components are affected?"
    allowed_params = [
        "cve_id",
        "affectedness",
        "affect_resolution",
        "affect_impact",
        "component_type",
        "namespace",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.cve_id = self.params.get("cve_id")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")
        self.component_type = self.params.get("component_type")
        self.namespace = self.params.get("namespace")

    def execute(self, status=None) -> dict:
        cond = {}
        if self.affectedness:
            cond["affectedness"] = self.affectedness
        if self.affect_resolution:
            cond["resolution"] = self.affect_resolution
        if self.affect_impact:
            cond["impact"] = self.affect_impact

        component_cond = {}
        if self.namespace:
            component_cond["namespace"] = self.namespace
        if self.component_type:
            component_cond["type"] = self.component_type

        flaw = self.osidb_session.flaws.retrieve(self.cve_id)
        affects = flaw.affects
        results = list()

        for affect in affects:
            components = self.corgi_session.components.retrieve_list_iterator_async(
                name=affect.ps_component,
                latest_components_by_streams="True",
                include_fields="purl,product_streams,product_versions,software_build",
            )
            for c in components:
                results.append(c.to_dict())

        return {
            "link": f"{OSIDB_SERVER_URL}/osidb/api/v1/flaws/{flaw.cve_id}",
            "cve_id": flaw.cve_id,
            "title": flaw.title,
            "description": flaw.description,
            "components": results,
        }


class cves_for_specific_component_query:
    """CVEs affecting a specific component?"""

    # include count
    name = "cves_for_specific component"
    description = "Which CVEs affect a specific component ?"
    allowed_params = [
        "component_name",
        "purl",
        "flaw_state",
        "flaw_impact",
        "flaw_resolution",
        "affectedness",
        "affect_resolution",
        "affect_impact",
        "strict_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.component_name = self.params.get("component_name")
        self.purl = self.params.get("purl")
        self.flaw_state = self.params.get("flaw_state")
        self.flaw_impact = self.params.get("flaw_impact")
        self.flaw_resolution = self.params.get("flaw_resolution")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        components = []
        if self.component_name:
            affects: list = []
            params = {
                "include_fields": "cve_id,title,resolution,impact,affects",
            }
            params["affects__ps_component"] = self.component_name
            if self.flaw_state:
                params["state"] = self.flaw_state
            if self.flaw_resolution:
                params["resolution"] = self.flaw_resolution
            if self.flaw_impact:
                params["impact"] = self.flaw_impact
            if self.affectedness:
                params["affects__affectedness"] = self.affectedness
            if self.affect_resolution:
                params["affects__resolution"] = self.affect_resolution
            if self.affect_impact:
                params["affects__impact"] = self.affect_impact

            flaws = self.osidb_session.flaws.retrieve_list_iterator_async(**params)

            for flaw in flaws:
                for affect in flaw.affects:
                    if self.affectedness:
                        if self.affectedness != affect.affectedness:
                            continue
                    if self.affect_resolution:
                        if self.affect_resolution != affect.resolution:
                            continue
                    if self.affect_impact:
                        if self.affect_impact != affect.impact:
                            continue
                    affects.append(
                        {
                            "link_affect": f"{OSIDB_SERVER_URL}/osidb/api/v1/affects/{affect.uuid}",  # noqa
                            "link_cve": f"{OSIDB_SERVER_URL}/osidb/api/v1/flaws/{flaw.cve_id}",  # noqa
                            "link_component": f"{CORGI_SERVER_URL}/api/v1/components?name={affect.ps_component}&latest_components_by_streams=True",  # noqa
                            "link_community_component": f"{COMMUNITY_COMPONENTS_SERVER_URL}/api/v1/components?name={affect.ps_component}&latest_components_by_streams=True",  # noqa
                            "flaw_cve_id": flaw.cve_id,
                            "title": flaw.title,
                            "flaw_resolution": flaw.resolution,
                            "affect_component_name": affect.ps_component,
                            "affect_product_version": affect.ps_module,
                            "affect_affectedness": affect.affectedness,
                            "affect_impact": affect.impact,
                            "affect_resolution": affect.resolution,
                        }
                    )
            components.append(
                {
                    "link": f"{CORGI_SERVER_URL}/api/v1/components?name={self.component_name}",
                    "name": self.component_name,
                    "affects": affects,
                }
            )

        if self.purl:
            pass

        return components


class cves_for_specific_product_query:
    name = "cves-for-product"
    description = "What cves affect a specific product ?"
    allowed_params = [
        "product_version_name",
        "ofuri",
        "flaw_state",
        "flaw_impact",
        "flaw_resolution",
        "affectedness",
        "affect_resolution",
        "affect_impact",
        "strict_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.product_version_name = self.params.get("product_version_name")
        self.ofuri = self.params.get("ofuri")
        self.flaw_state = self.params.get("flaw_state")
        self.flaw_impact = self.params.get("flaw_impact")
        self.flaw_resolution = self.params.get("flaw_resolution")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        components = []
        if self.product_version_name:
            affects: list = []
            params = {
                "include_fields": "cve_id,title,resolution,impact,affects",
            }
            params["affects__ps_module"] = self.product_version_name
            if self.flaw_resolution:
                params["resolution"] = self.flaw_resolution
            if self.flaw_impact:
                params["impact"] = self.flaw_impact
            if self.affectedness:
                params["affects__affectedness"] = self.affectedness
            if self.affect_resolution:
                params["affects__resolution"] = self.affect_resolution
            if self.affect_impact:
                params["affects__impact"] = self.affect_impact

            flaws = self.osidb_session.flaws.retrieve_list_iterator_async(**params)
            for flaw in flaws:
                for affect in flaw.affects:
                    if self.affectedness:
                        if self.affectedness != affect.affectedness:
                            continue
                    if self.affect_resolution:
                        if self.affect_resolution != affect.resolution:
                            continue
                    if self.affect_impact:
                        if self.affect_impact != affect.impact:
                            continue
                    affects.append(
                        {
                            "link_affect": f"{OSIDB_SERVER_URL}/osidb/api/v1/affects/{affect.uuid}",  # noqa
                            "link_cve": f"{OSIDB_SERVER_URL}/osidb/api/v1/flaws/{flaw.cve_id}",  # noqa
                            "link_component": f"{CORGI_SERVER_URL}/api/v1/components?name={affect.ps_component}&latest_components_by_streams=True",  # noqa
                            "link_community_component": f"{COMMUNITY_COMPONENTS_SERVER_URL}/api/v1/components?name={affect.ps_component}&latest_components_by_streams=True",  # noqa
                            "flaw_cve_id": flaw.cve_id,
                            "title": flaw.title,
                            "flaw_state": flaw.state,
                            "flaw_resolution": flaw.resolution,
                            "affect_component_name": affect.ps_component,
                            "affect_product_version": affect.ps_module,
                            "affect_affectedness": affect.affectedness,
                            "affect_impact": affect.impact,
                            "affect_resolution": affect.resolution,
                        }
                    )
            components.append(
                {
                    "link": f"{CORGI_SERVER_URL}/api/v1/product_versions?name={self.product_version_name}",  # noqa
                    "name": self.product_version_name,
                    "affects": affects,
                }
            )

        return components



"""
Griffon exceptions and exception handling

TODO: eventually we will want to gracefully handle all errors
though useful during development to 'see' everything.
"""

from functools import partial, wraps

import click

from griffon.helpers import Color, Style


class GriffonException(click.ClickException):
    """Base Griffon exception"""

    def __init__(self, message, *args, **kwargs) -> None:
        super().__init__(f"{Style.BOLD}{Color.RED}{message}{Color.RESET}", *args, **kwargs)

    def show(self, *args, **kwargs):
        click.echo((f"{Style.BOLD}{Color.RED}{' Griffon Error ':*^55}{Style.RESET}"))
        super().show(*args, **kwargs)


class GriffonUsageError(GriffonException, click.UsageError):
    """Griffon exception that signals a usage error"""

    pass


def catch_exception(func=None, *, handle):
    """catch exception decorator"""
    if not func:
        return partial(catch_exception, handle=handle)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except handle as e:
            raise click.GriffonException(e)

    return wrapper



"""
osidb entity operations

"""
import logging

import click

from .affects import affects
from .flaws import flaws
from .manage import manage_grp
from .trackers import trackers

logger = logging.getLogger("griffon")

default_conditions: dict = {}


@click.group(name="osidb")
@click.pass_context
def osidb_grp(ctx):
    pass


osidb_grp.add_command(flaws)
osidb_grp.add_command(affects)
osidb_grp.add_command(trackers)
osidb_grp.add_command(manage_grp)



"""
osidb tracker entities operations
"""

import logging

import click
from osidb_bindings.bindings.python_client.api.osidb import (
    osidb_api_v1_trackers_list,
    osidb_api_v1_trackers_retrieve,
)
from osidb_bindings.bindings.python_client.models import Tracker

from griffon import OSIDB_SERVER_URL, OSIDBService, progress_bar
from griffon.commands.entities.helpers import (
    multivalue_params_to_csv,
    query_params_options,
)
from griffon.output import cprint

logger = logging.getLogger("griffon")


@click.group(help=f"{OSIDB_SERVER_URL}/osidb/api/v1/trackers")
@click.pass_context
def trackers(ctx):
    """OSIDB Trackers."""
    pass


@trackers.command(name="list")
@query_params_options(
    entity="Tracker",
    endpoint_module=osidb_api_v1_trackers_list,
    options_overrides={
        "include_fields": {"type": click.Choice(OSIDBService.get_fields(Tracker))},
        "exclude_fields": {"type": click.Choice(OSIDBService.get_fields(Tracker))},
        "include_meta_attr": {"type": click.Choice(OSIDBService.get_meta_attr_fields(Tracker))},
    },
)
@click.pass_context
def list_trackers(ctx, **params):
    # TODO: handle pagination
    # TODO: handle output
    session = OSIDBService.create_session()

    params = multivalue_params_to_csv(params)
    return cprint(
        list(session.trackers.retrieve_list_iterator_async(max_results=5000, **params)), ctx=ctx
    )


@trackers.command(name="get")
@click.option("--uuid", "tracker_uuid", help="Tracker UUID.", required=True)
@query_params_options(
    entity="Tracker",
    endpoint_module=osidb_api_v1_trackers_retrieve,
    options_overrides={
        "include_fields": {"type": click.Choice(OSIDBService.get_fields(Tracker))},
        "exclude_fields": {"type": click.Choice(OSIDBService.get_fields(Tracker))},
        "include_meta_attr": {"type": click.Choice(OSIDBService.get_meta_attr_fields(Tracker))},
    },
)
@click.pass_context
@progress_bar
def get_tracker(ctx, tracker_uuid, **params):
    params = multivalue_params_to_csv(params)

    session = OSIDBService.create_session()
    data = session.trackers.retrieve(tracker_uuid, **params)
    return cprint(data, ctx=ctx)



"""
osidb flaw comment entities operations
"""

import json
import logging

import click
from osidb_bindings.bindings.python_client.api.osidb import (
    osidb_api_v1_flaws_comments_create,
    osidb_api_v1_flaws_comments_list,
    osidb_api_v1_flaws_comments_retrieve,
)
from osidb_bindings.bindings.python_client.models import FlawComment
from requests import HTTPError

from griffon import OSIDB_SERVER_URL, OSIDBService, progress_bar
from griffon.commands.entities.helpers import (
    filter_request_fields,
    get_editor,
    multivalue_params_to_csv,
    query_params_options,
    request_body_options,
)
from griffon.exceptions import GriffonException
from griffon.output import console, cprint

logger = logging.getLogger("griffon")


@click.group(help=f"{OSIDB_SERVER_URL}/osidb/api/v1/flaws/<id>/comments", name="comments")
@click.pass_context
def flaw_comments(ctx):
    """OSIDB Flaw Comments."""
    pass


@flaw_comments.command(name="get")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@click.option("--uuid", "comment_uuid", help="Comment UUID.", required=True)
@query_params_options(
    entity="Flaw Comment",
    endpoint_module=osidb_api_v1_flaws_comments_retrieve,
    options_overrides={
        "include_fields": {"type": click.Choice(OSIDBService.get_fields(FlawComment))},
        "exclude_fields": {"type": click.Choice(OSIDBService.get_fields(FlawComment))},
        "include_meta_attr": {"type": click.Choice(OSIDBService.get_meta_attr_fields(FlawComment))},
    },
)
@click.pass_context
@progress_bar
def get_flaw_comment(ctx, flaw_id, comment_uuid, **params):
    params = multivalue_params_to_csv(params)

    session = OSIDBService.create_session()
    data = session.flaws.comments.retrieve(flaw_id, comment_uuid, **params)
    return cprint(data, ctx=ctx)


@flaw_comments.command(name="list")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@query_params_options(
    entity="Flaw Comments",
    endpoint_module=osidb_api_v1_flaws_comments_list,
    options_overrides={
        "include_fields": {"type": click.Choice(OSIDBService.get_fields(FlawComment))},
        "exclude_fields": {"type": click.Choice(OSIDBService.get_fields(FlawComment))},
        "include_meta_attr": {"type": click.Choice(OSIDBService.get_meta_attr_fields(FlawComment))},
    },
)
@click.pass_context
@progress_bar
def list_flaw_comments(ctx, flaw_id, **params):
    # TODO: handle pagination
    # TODO: handle output
    session = OSIDBService.create_session()

    params = multivalue_params_to_csv(params)
    data = session.flaws.comments.retrieve_list(flaw_id, **params).results
    return cprint(data, ctx=ctx)


@flaw_comments.command(name="create")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@request_body_options(
    endpoint_module=osidb_api_v1_flaws_comments_create,
    exclude=["uuid", "created_dt"],
)
@click.pass_context
def create_flaw_comment(ctx, flaw_id, **params):
    request_body_type = getattr(osidb_api_v1_flaws_comments_create, "REQUEST_BODY_TYPE", None)
    if request_body_type is None:
        raise GriffonException(
            "No request body template for Flaw Comment create. "
            "Is correct version of osidb-bindings installed?"
        )

    fields = filter_request_fields(
        request_body_type.get_fields(),
        exclude=["uuid", "created_dt"],
    )
    params = multivalue_params_to_csv(params)

    session = OSIDBService.create_session()

    data = {field: "" for field in fields}
    data.update((field, value) for field, value in params.items() if value is not None)

    if ctx.obj["EDITOR"]:
        data = click.edit(
            text=json.dumps(data, indent=4, default=str), editor=get_editor(), require_save=False
        )
        data = json.loads(data)

    try:
        data = session.flaws.comments.create(data, flaw_id=flaw_id)
    except HTTPError as e:
        if ctx.obj["VERBOSE"]:
            console.log(e, e.response.json())
        raise GriffonException(
            "Failed to create Flaw Comment. "
            "You might have insufficient permission or you've supplied malformed data. "
            "Consider running griffon with -v option for verbose error log."
        )
    return cprint(data, ctx=ctx)



"""
osidb manage operations

"""
import logging

import click

from griffon import OSIDB_SERVER_URL, OSIDBService
from griffon.output import cprint

logger = logging.getLogger("griffon")

default_conditions: dict = {}


@click.group(name="admin")
@click.pass_context
def manage_grp(ctx):
    """Manage osidb"""
    pass


@manage_grp.command(name="status")
@click.pass_context
def osidb_status(ctx):
    session = OSIDBService.create_session()
    data = session.status()
    return cprint(data, ctx=ctx)


@manage_grp.command(name="api_doc")
def osidb_api_docs():
    click.launch(f"{OSIDB_SERVER_URL}/osidb/api/v1/schema/swagger-ui/")



"""
osidb flaw package version entities operations
"""

import json
import logging

import click
from osidb_bindings.bindings.python_client.api.osidb import (
    osidb_api_v1_flaws_package_versions_create,
    osidb_api_v1_flaws_package_versions_list,
    osidb_api_v1_flaws_package_versions_retrieve,
    osidb_api_v1_flaws_package_versions_update,
)
from osidb_bindings.bindings.python_client.models import FlawPackageVersion
from requests import HTTPError

from griffon import OSIDB_SERVER_URL, OSIDBService, progress_bar
from griffon.commands.entities.helpers import (
    abort_if_false,
    filter_request_fields,
    get_editor,
    multivalue_params_to_csv,
    query_params_options,
    request_body_options,
)
from griffon.exceptions import GriffonException
from griffon.output import console, cprint

logger = logging.getLogger("griffon")


@click.group(
    help=f"{OSIDB_SERVER_URL}/osidb/api/v1/flaws/<id>/package_versions", name="package_versions"
)
@click.pass_context
def flaw_package_versions(ctx):
    """OSIDB Flaw Package Versions."""
    pass


@flaw_package_versions.command(name="get")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@click.option("--uuid", "package_version_uuid", help="Package Version UUID.", required=True)
@query_params_options(
    entity="Flaw Package Version",
    endpoint_module=osidb_api_v1_flaws_package_versions_retrieve,
    options_overrides={
        "include_fields": {"type": click.Choice(OSIDBService.get_fields(FlawPackageVersion))},
        "exclude_fields": {"type": click.Choice(OSIDBService.get_fields(FlawPackageVersion))},
        "include_meta_attr": {
            "type": click.Choice(OSIDBService.get_meta_attr_fields(FlawPackageVersion))
        },
    },
)
@click.pass_context
@progress_bar
def get_flaw_package_version(ctx, flaw_id, package_version_uuid, **params):
    params = multivalue_params_to_csv(params)

    session = OSIDBService.create_session()
    data = session.flaws.package_versions.retrieve(flaw_id, package_version_uuid, **params)
    return cprint(data, ctx=ctx)


@flaw_package_versions.command(name="list")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@query_params_options(
    entity="Flaw Package Version",
    endpoint_module=osidb_api_v1_flaws_package_versions_list,
    options_overrides={
        "include_fields": {"type": click.Choice(OSIDBService.get_fields(FlawPackageVersion))},
        "exclude_fields": {"type": click.Choice(OSIDBService.get_fields(FlawPackageVersion))},
        "include_meta_attr": {
            "type": click.Choice(OSIDBService.get_meta_attr_fields(FlawPackageVersion))
        },
    },
)
@click.pass_context
@progress_bar
def list_flaw_package_versions(ctx, flaw_id, **params):
    # TODO: handle pagination
    # TODO: handle output
    session = OSIDBService.create_session()

    params = multivalue_params_to_csv(params)
    data = session.flaws.package_versions.retrieve_list(flaw_id, **params).results
    return cprint(data, ctx=ctx)


@flaw_package_versions.command(name="create")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@request_body_options(
    endpoint_module=osidb_api_v1_flaws_package_versions_create,
    exclude=["uuid", "created_dt"],
)
@click.pass_context
def create_flaw_package_version(ctx, flaw_id, **params):
    request_body_type = getattr(
        osidb_api_v1_flaws_package_versions_create, "REQUEST_BODY_TYPE", None
    )
    if request_body_type is None:
        raise GriffonException(
            "No request body template for Flaw Package Version create. "
            "Is correct version of osidb-bindings installed?"
        )

    fields = filter_request_fields(
        request_body_type.get_fields(),
        exclude=["uuid", "created_dt"],
    )
    params = multivalue_params_to_csv(params)

    session = OSIDBService.create_session()

    data = {field: "" for field in fields}
    data.update((field, value) for field, value in params.items() if value is not None)

    if ctx.obj["EDITOR"]:
        data = click.edit(
            text=json.dumps(data, indent=4, default=str), editor=get_editor(), require_save=False
        )
        data = json.loads(data)

    try:
        data = session.flaws.package_versions.create(data, flaw_id=flaw_id)
    except HTTPError as e:
        if ctx.obj["VERBOSE"]:
            console.log(e, e.response.json())
        raise GriffonException(
            "Failed to create Flaw Package Version. "
            "You might have insufficient permission or you've supplied malformed data. "
            "Consider running griffon with -v option for verbose error log."
        )
    return cprint(data, ctx=ctx)


@flaw_package_versions.command(name="update")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@click.option("--uuid", "package_version_uuid", help="Package Version UUID.", required=True)
@request_body_options(endpoint_module=osidb_api_v1_flaws_package_versions_update, exclude=["uuid"])
@click.pass_context
def update_flaw_package_version(ctx, flaw_id, package_version_uuid, **params):
    request_body_type = getattr(
        osidb_api_v1_flaws_package_versions_update, "REQUEST_BODY_TYPE", None
    )
    if request_body_type is None:
        raise GriffonException(
            "No request body template for Flaw Package Version update. "
            "Is correct version of osidb-bindings installed?"
        )

    fields = filter_request_fields(request_body_type.get_fields(), exclude=["uuid"])
    params = multivalue_params_to_csv(params)

    session = OSIDBService.create_session()

    try:
        data = session.flaws.package_versions.retrieve(
            flaw_id, package_version_uuid, include_fields=",".join(fields)
        )
    except Exception as e:
        if ctx.obj["VERBOSE"]:
            console.log(e, e.response.json())
        raise GriffonException(
            f"Failed to fetch Flaw Package Version with ID '{package_version_uuid}'. "
            "Flaw or Flaw Package Version either does not exist or you have "
            "insufficient permissions. "
            "Consider running griffon with -v option for verbose error log."
        )

    data = data.to_dict()
    # remove status data from OSIDB server
    [data.pop(key) for key in ["dt", "env", "revision", "version"]]
    data.update((field, value) for field, value in params.items() if value is not None)

    if ctx.obj["EDITOR"]:
        data = click.edit(text=json.dumps(data, indent=4), editor=get_editor(), require_save=False)
        data = json.loads(data)

    try:
        data = session.flaws.package_versions.update(flaw_id, data, package_version_uuid)
    except HTTPError as e:
        if ctx.obj["VERBOSE"]:
            console.log(e, e.response.json())
        raise GriffonException(
            f"Failed to update Flaw Package Version with ID '{package_version_uuid}'. "
            "You might have insufficient permission or you've supplied malformed data. "
            "Consider running griffon with -v option for verbose error log."
        )
    return cprint(data, ctx=ctx)


@flaw_package_versions.command(name="delete")
@click.option(
    "--flaw-id",
    "flaw_id",
    help="Flaw CVE-ID or UUID.",
    required=True,
)
@click.option("--uuid", "package_version_uuid", help="Package Version UUID.", required=True)
@click.option(
    "--yes",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to delete Flaw Package Version?",
)
@click.pass_context
def delete_flaw_package_version(ctx, flaw_id, package_versions_uuid, **params):
    session = OSIDBService.create_session()
    try:
        data = session.flaws.package_versions.delete(flaw_id, package_versions_uuid)
    except HTTPError as e:
        if ctx.obj["VERBOSE"]:
            console.log(e, e.response.json())
        raise GriffonException(
            f"Failed to delete Flaw Package Version {package_versions_uuid}. "
            "It either does not exist or you have insufficient permissions."
        )
    return cprint(data, ctx=ctx)
